@@ -1,49 +1,46 @@
 """Main module."""
 
 import re
-from asyncio import wait_for
-from asyncio import get_event_loop
-from random import randint
-from codecs import lookup 
+from asyncio import get_event_loop, wait_for
+from codecs import lookup
 from functools import partial
-from json import dumps
-from json import loads
-from ssl import SSLContext
 from gzip import decompress as gzip_decompress
+from http import cookies
+from io import IOBase
+from json import dumps, loads
+from os.path import basename
+from random import randint
+from ssl import SSLContext
+from typing import (
+    Any,
+    AsyncIterator,
+    Callable,
+    Dict,
+    Iterator,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
+from urllib.parse import ParseResult, urlencode, urlparse
 from zlib import decompress as zlib_decompress
 
-from io import IOBase
-from os.path import basename
-from urllib.parse import urlparse
-from urllib.parse import urlencode
-from urllib.parse import ParseResult
-
-from typing import Any
-from typing import AsyncIterator
-from typing import Callable
-from typing import Dict
-from typing import Iterator
-from typing import Union
-from typing import Optional
-from typing import List
-from typing import Tuple
-
 import chardet
-
-from aiosonic_utils.structures import CaseInsensitiveDict
-
 from aiosonic.connection import Connection
-from aiosonic.exceptions import ConnectTimeout
-from aiosonic.exceptions import HttpParsingError
-from aiosonic.exceptions import MaxRedirects
-from aiosonic.exceptions import MissingWriterException
-from aiosonic.exceptions import ReadTimeout
-from aiosonic.exceptions import RequestTimeout
 from aiosonic.connectors import TCPConnector
-from aiosonic.exceptions import TimeoutException
+from aiosonic.exceptions import (
+    ConnectTimeout,
+    HttpParsingError,
+    MaxRedirects,
+    MissingWriterException,
+    ReadTimeout,
+    RequestTimeout,
+    TimeoutException,
+)
 from aiosonic.timeout import Timeouts
 from aiosonic.utils import cache_decorator
 from aiosonic.version import VERSION
+from aiosonic_utils.structures import CaseInsensitiveDict
 
 # TYPES
 from aiosonic.types import ParamsType
@@ -92,27 +89,53 @@ class HttpHeaders(CaseInsensitiveDict):
 HeadersType = Union[Dict[str, str], List[Tuple[str, str]], HttpHeaders]
 
 
-def _add_header(headers: HeadersType, key: str, value: str):
+def _headers_iterator(headers: HeadersType):
+    iterator = headers if isinstance(headers, List) else headers.items()
+    for key, data in iterator:
+        yield key, data
+
+
+def _add_headers(headers: HeadersType, headers_to_add: HeadersType):
+    """Safe add multiple headers."""
+    for key, data in _headers_iterator(headers_to_add):
+        _add_header(headers, key, data)
+
+
+def _add_header(headers: HeadersType, key: str, value: str, replace=False):
     """Safe add header method."""
     if isinstance(headers, List):
-        included = [item for item in headers if item[0] == key]
-        if included:
-            headers.remove(included[0])
+        if replace:
+            included = [item for item in headers if item[0] == key]
+            if included:
+                headers.remove(included[0])
         headers.append((key, value))
     else:
         headers[key] = value
 
 
+async def _parse_headers_iterator(connection: Connection):
+    """Transform loop to iterator."""
+    while True:
+        res_data = await connection.reader.readline()
+        if b': ' not in res_data and b':' not in res_data:
+            break
+        yield res_data
+
+
 class HttpResponse:
     """Custom HttpResponse class for handling responses.
 
     Properties:
-      * status_code (int): response status code
-      * headers (HttpHeaders): headers in case insensitive dict
-      * raw_headers (List[Tuple[bytes, bytes]]): headers as raw format
+      * **status_code** (int): response status code
+      * **headers** (:class:`aiosonic.HttpHeaders`): headers in case insensitive dict
+      * **cookies** (:class:`http.cookies.SimpleCookie`): instance of SimpleCookies
+        if cookies present in respone.
+      * **raw_headers** (List[Tuple[bytes, bytes]]): headers as raw format
     """
+
     def __init__(self):
         self.headers = HttpHeaders()
+        self.cookies = None
         self.raw_headers = []
         self.body = b''
         self.response_initial = None
@@ -134,6 +157,20 @@ class HttpResponse:
         self.headers[key] = val
         self.raw_headers.append((key, val))
 
+    async def _set_response_headers(self, iterator):
+        async for header_data in iterator:
+            header_tuple = HttpHeaders._clear_line(header_data)
+            self._set_header(*header_tuple)
+
+            # set cookies in response
+            if header_tuple[0].lower() == b'set-cookie':
+                self._update_cookies(header_tuple)
+
+    def _update_cookies(self, header_tuple):
+        """Update jar of cookies."""
+        self.cookies = self.cookies or cookies.SimpleCookie()
+        self.cookies.load(header_tuple[1].decode())
+
     def _set_connection(self, connection: Connection):
         """Set header to response."""
         self.connection = connection
@@ -232,14 +269,14 @@ def _get_header_data(url: ParseResult,
     get_base = f'{uppercase_method} {path} HTTP/1.1{_NEW_LINE}'
 
     port = url.port or (443 if url.scheme == 'https' else 80)
-    hostname = url.hostname
+    hostname = str(url.hostname)
 
     if port != 80:
         hostname += ':' + str(port)
 
-    headers_base = {}
+    headers_base = []
     if http2conn:
-        headers_base.update({
+        _add_headers(headers_base, {
             ':method': method,
             ':authority': hostname.split(':')[0],
             ':scheme': 'https',
@@ -247,22 +284,23 @@ def _get_header_data(url: ParseResult,
             'user-agent': f'aioload/{VERSION}'
         })
     else:
-        headers_base.update({
+        _add_headers(headers_base, {
             'HOST': hostname,
             'Connection': 'keep-alive',
             'User-Agent': f'aioload/{VERSION}'
         })
 
     if multipart:
-        headers_base['Content-Type'] = f'multipart/form-data; boundary="{boundary}"'
+        _add_header(headers_base, 'Content-Type',
+                    f'multipart/form-data; boundary="{boundary}"')
 
     if headers:
-        headers_base.update(headers)
+        _add_headers(headers_base, headers)
 
     if http2conn:
         return headers_base
 
-    for key, data in headers_base.items():
+    for key, data in _headers_iterator(headers_base):
         get_base += f'{key}: {data}{_NEW_LINE}'
     return (get_base + _NEW_LINE).encode()
 
@@ -352,7 +390,8 @@ async def _send_multipart(data: Dict[str, str],
             val.close()
 
         else:
-            to_send += (f'Content-Disposition: form-data; name="{key}"{_NEW_LINE}{_NEW_LINE}').encode()
+            to_send += (
+                f'Content-Disposition: form-data; name="{key}"{_NEW_LINE}{_NEW_LINE}').encode()
             to_send += val.encode() + _NEW_LINE.encode()
 
     # write --boundary-- for finish
@@ -400,11 +439,8 @@ async def _do_request(urlparsed: ParseResult,
 
         res_data = None
         # reading headers
-        while True:
-            res_data = await connection.reader.readline()
-            if b': ' not in res_data and b':' not in res_data:
-                break
-            response._set_header(*HttpHeaders._clear_line(res_data))
+
+        await response._set_response_headers(_parse_headers_iterator(connection))
 
         size = response.headers.get(b'content-length')
         chunked = response.headers.get(b'transfer-encoding', '') == b'chunked'
@@ -428,14 +464,28 @@ class HTTPClient:
     """aiosonic.HTTPClient class.
 
     This class holds the client creation that will be used for requests.
-    """
-    def __init__(self, connector: TCPConnector = None):
-        """Initialize client options.
 
-        Params:
-            * **connector**: TCPConnector to be used if provided
-        """
+    Params:
+        * **connector**: TCPConnector to be used if provided
+        * **handle_cookies**: Flag to indicate if keep response cookies in
+            client and send them in next requests.
+    """
+
+    def __init__(self, connector: TCPConnector = None, handle_cookies=False):
+        """Initialize client options."""
         self.connector = connector or TCPConnector()
+        self.handle_cookies = handle_cookies
+        self.cookies_map: Dict[str, cookies.SimpleCookie] = {}
+
+    async def __aenter__(self):
+        return self
+
+    async def __aexit__(self, *_args, **_kwargs):
+        await self.shutdown()
+
+    async def shutdown(self):
+        """Cleanup connections, this method makes client unusable."""
+        await self.connector.cleanup()
 
     async def _request_with_body(self,
                                  url: str,
@@ -634,9 +684,12 @@ class HTTPClient:
         urlparsed = _get_url_parsed(url)
 
         boundary = None
-        headers = headers or {}
+        headers = headers or []
         body: ParsedBodyType = b''
 
+        if self.handle_cookies:
+            self._add_cookies_to_request(urlparsed.hostname, headers)
+
         if method != 'GET' and data and not multipart:
             body = _setup_body_request(data, headers)
         elif multipart:
@@ -661,12 +714,19 @@ class HTTPClient:
                     timeout=(timeouts
                              or self.connector.timeouts).request_timeout)
 
+                if self.handle_cookies:
+                    self._save_new_cookies(urlparsed.hostname, response)
+
                 if follow and response.status_code in {301, 302}:
                     max_redirects -= 1
 
                     if max_redirects == 0:
                         raise MaxRedirects()
 
+                    if self.handle_cookies:
+                        self._add_cookies_to_request(
+                            urlparsed.hostname, headers)
+
                     parsed_full_url = _get_url_parsed(
                         response.headers[b'location'].decode())
 
@@ -697,3 +757,18 @@ class HTTPClient:
                 self.connector.wait_free_pool(), timeout)
         except TimeoutException:
             return False
+
+    def _add_cookies_to_request(self, host: str, headers: HeadersType):
+        """Add cookies to request."""
+        host_cookies = self.cookies_map.get(host)
+        if (host_cookies and
+                not any([header.lower() == 'cookie'
+                         for header, _ in headers])):
+            cookies_str = host_cookies.output(header='Cookie:')
+            for cookie_data in cookies_str.split('\r\n'):
+                _add_header(headers, *cookie_data.split(': ', 1))
+
+    def _save_new_cookies(self, host: str, response: HttpResponse):
+        """Save new cookies in map."""
+        if response.cookies:
+            self.cookies_map[host] = response.cookies
